#|[
Author: Juergen Lorenz
ju (at) jugilo (dot) de

Copyright (c) 2015-2016, Juergen Lorenz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of the author nor the names of its contributors may be
used to endorse or promote products derived from this software without
specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]|#

;;; This module exports only one macro, local, which is inspired by the
;;; equally named operator in ML. It exports a list of definitions,
;;; depending on another hidden list of definitions. So it can be used
;;; to implement procedures depending on a common state, something what
;;; define-values can do as well.

(module locals *
  (import scheme)
  (import (only chicken print))
  (import-for-syntax
    (only chicken assert)
    (only data-structures disjoin list-of?))

;;; (local ((hidden hxpr)|
;;;         ((hidden . hargs) . hbody)
;;;         ...)
;;;   (exported expr)|
;;;   ((exported . eargs) . ebody)
;;;   ....)
;;; --------------------------------------------------
;;; translates into
;;; (begin
;;;   (define exported #f)
;;;   ....
;;;   (letrec ((hidden hxpr)|
;;;            (hidden (lambda hargs . hbody))
;;;            ...)
;;;     (set! exported expr)|
;;;     (set! exported (lambda eargs . ebody)
;;;     ....)))
(define-syntax local
  (ir-macro-transformer
    (lambda (form inject compare?)
      (let ((locals (cadr form))
            (defs (cddr form))
            (pair-of-symbol?
              (lambda (xpr)
                (let loop ((xpr xpr))
                  (or (symbol? xpr)
                      (and (pair? xpr)
                           (symbol? (car xpr))
                           (loop (cdr xpr))))))))
        (assert ((list-of? pair?) locals))
        (assert ((list-of?
                   (disjoin pair-of-symbol?
                            (list-of? symbol?)))
                 (map car locals)))
        (assert (pair? defs))
        (assert ((list-of? pair?) defs))
        (assert ((list-of?
                   (disjoin pair-of-symbol?
                            (list-of? symbol?)))
                 (map car defs)))
        (let ((pair->var (lambda (p)
                           (if (pair? (car p)) ; procedure
                             (caar p)
                             (car p))))
              (pair->val (lambda (p)
                           (if (pair? (car p)) ; procedure
                             `(lambda ,(cdar p)
                                ,@(cdr p))
                             (cadr p)))))
          (let ((local-vars (map pair->var locals))
                (local-vals (map pair->val locals))
                (vars (map pair->var defs))
                (vals (map pair->val defs)))
            `(begin
               ,@(map (lambda (n)
                        `(define ,n #f))
                      vars)
               (letrec ,(map (lambda (n l)
                               `(,n ,l))
                             local-vars local-vals)
                 ,@(map (lambda (n l)
                          `(set! ,n ,l))
                        vars vals)))))))))

(define (locals)
  (for-each print
    (list "local"
          "macro:"
          "  (local ((hidden hxpr)|((hidden . hargs) . hbody)"
          "          ...)"
          "    (exported expr)|((exported . eargs) . ebody)"
          "    ....)"
          "exports some definitions exported ...."
          "depending on other definitions hidden ..."
          "which are hidden")))

  ) ; module locals
